/*
  Macro          HistRepair
  Author         Carlo Hogeveen
  Website        eCarlo.nl/tse
  Version        v0.1   17 Jul 2023


  THIS IS A BETA VERSION !!!

  It has been tested on my systems and there it has no known errors.
  Please report any errors or imperfections.



  This tool and extension repairs any errors in the tsehist.dat file.
  This is the file in which TSE stores its history lists.

  You can execute the macro as a tool and/or add it to the Macro AutoLoad List
  as an extension.

  The tool/extension's input is the tsehist.dat file from before TSE started.
  If it finds no errors, then it does nothing.
  if it finds errors, then it asks you to accept or reject repairing them.
  If you accept, then it postpones the repair to after TSE closes,
  because it needs to wait because TSE writes tsehist.dat when it closes.
  When TSE closes, it again asks you to confirm or abort the repair.
  if you confirm, then the repair is done a second after TSE closes.
  The repair undoes the normal TSE history changes of the closed TSE session.



  CONTEXT DETAILS

  TSE remembers it history lists in a tsehist.dat file.
  TSE recreates this file every time TSE closes.
  The file can contain errors caused by bugs in TSE and macros.
  Not all of these errors can be corrected during a TSE session.

  All history lists are identified by a number from 1 through 255.
  There are two types of history list:
  - A "fixed" history list is one that is built in into TSE.
    It has a fixed number from 128 through 255, most of which are unused.
    It has a built-in "name" that is actually a synonym for its number.
    See TSE's Help for AddHistoryStr() for more info.
    See the "Datadef fixed_history_synonyms" in the source code below
    for which fixed history list numbers have which synonyms.
  - A "free" history list is assigned a number from 1 through 127 by a macro
    that uses GetFreeHistory() to get that number based on a freely
    choosable real name.
    In theory GetFreeHistory() either always returns the same number for the
    same name, or it returns 0 to indicate failure to do so.

  In practice there are 3 general problems with free history lists:
  - There can only be 127, so when heavily used they run out.
  - GetFreeHistory() itself has bugs (17 Jul 2023).
  - Macros do not check if GetFreeHistory() succeeds.   :-(
    Currently (17 Jul 2023) all Semware's and my macros that use free history
    lists all have this fault.
    If GetFreeHistory() fails then they issue no error and blindly continue.
    I think this bad programming habit was established in earlier days when
    using up all 127 free history lists seemed unreachable.

  Unsurprisingly, my tsehist.dat was a mess. Yours might be too.
  This tool aims to repair tsehist.dat.



  TOOL DETAILS

  This repair tool works in two stages:
  - Stage 1:
    When the tool is loaded or executed it:
    - Prepares actions for a new tsehist.dat file:
      - Deleting empty file lines. (Not the same as empty history lines.)
      - Deleteing "names" for fixed history lists.
      - Deleting free history 0.
      - Deleting empty free history lists.             (Names without values.)
      - Deleting free list values without a list name. (Values without names.)
        An empty name is allowed and valid.
      - Deleting other names or whole list for lists with multiple names.
        The user is asked which action to perform.
    - Lets the user accept or reject the actions.
    - If the actions are accepted:
      - Creates a backup of tsehist.dat as tsehist_yyyymmdd_hhmmss.dat
        in the same directory using the file's old date and time.
      - Creates a new updated file tsehist_yyyymmdd_hhmmss.dat
        in the same directory using a current date and time.
  - Stage 2:
    After TSE closed and TSE wrote the tsehist.dat file,
    the tool overwrites the tsehist.dat file with the prepared one.
    Side-effect:
      This deletes any new TSE history changes that occurred the normal way
      during the closed TSE session.

  The tool can be executed as a macro and/or added to the Macro AutoLoad List.



  HISTORY

  v0.0.0.1   10 Jul 2023
    Initial "in development" release.
    The purpose of this release is to pre-inform Semware,
    who are working on other sides of the same problem.

  v0.0.0.2   13 Jul 2023
    Done:
      Extra error checks.
      More condensed error reporting.
      It now also works as an extension:
        It is ready enough, that I added it to my Macro AutoLoad List.
      It now updates tsehist.dat after two user confirmations:
        When it starts and when TSE closes.
      It creates its own backups of tsehist.dat, so what could go wrong?
    Todo:
      More testing.
      Some finishing touches.
      It is currently only compatible with a recent Windows GUI version of TSE.

  v0.0.0.3   15 Jul 2023
    It now maintains a cumulative "tsehist_HistRepair.log" history of errors.

    It now distinguishes the filenames of tsehist.dat's backup and new version
    by adding a "_bak" and "_new" filename part respectively.

  v0.1   17 Jul 2023
    Made the tool compatible with TSE's Console and Linux variants.
    Made the tool backwards compatible with TSE v4.0 upwards.

*/



// Start of compatibility restrictions and mitigations.


#ifdef LINUX
  #define WIN32 FALSE
#endif

#ifdef WIN32
#else
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif


#ifdef LINUX
  string COPY_COMMAND          [8] = 'cp -p -f'
  string COPY_OPTION           [1] = ''
  string DELETE_COMMAND        [5] = 'rm -f'
  string DELETE_OPTION         [1] = ''
  string NEWLINE               [1] = Chr(10)
  string PING_FREQUENCY_OPTION [1] = 'c'
  string SLASH                 [1] = '/'
#else
  string COPY_COMMAND          [5] = 'copy'
  string COPY_OPTION           [3] = ' /y'
  string DELETE_COMMAND        [5] = 'del'
  string DELETE_OPTION         [3] = ' /f'
  string NEWLINE               [2] = Chr(13) + Chr(10)
  string PING_FREQUENCY_OPTION [1] = 'n'
  string SLASH                 [1] = '\'
#endif


#if EDITOR_VERSION >= 4400h
  #define NEW_TIMEFORMAT 3
#else
  #define NEW_TIMEFORMAT 1
#endif


integer proc copy_file(string src_file, string dst_file)
  integer result = FALSE

  #if EDITOR_VERSION >= 4200h
    // Note: The oldest Linux TSE version was a TSE 4.2 beta version.
    result = CopyFile(src_file, dst_file)
  #else
    integer command_return_code = 0
    integer dos_errorlevel      = 0
    integer dos_io_result       = 0
    integer dos_result          = 0

    dos_result          = Dos(Format('copy'; QuotePath(src_file);
                                     QuotePath(dst_file); '/y'),
                              DOS_SYNC_CALL_FLAGS)
    dos_io_result       = DosIOResult()
    command_return_code = HiByte(dos_io_result)
    dos_errorlevel      = LoByte(dos_io_result)

    if      dos_result
    and not command_return_code
    and not dos_errorlevel
      result = TRUE
    endif
  #endif

  return(result)
end copy_file


// End of compatibility restrictions and mitigations.



// Global constants and semi-constants
#define DOS_ASYNC_CALL_FLAGS _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_|_RUN_DETACHED_|_DONT_WAIT_
#define DOS_SYNC_CALL_FLAGS  _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_
string  MACRO_NAME [MAXSTRINGLEN] = ''


Datadef fixed_history_synonyms
  '128 _EDIT_HISTORY_'
  '136 _NEWNAME_HISTORY_'
  '137 _EXECMACRO_HISTORY_'
  '138 _LOADMACRO_HISTORY_'
  '139 _KEYMACRO_HISTORY_'
  '144 _GOTOLINE_HISTORY_'
  '145 _GOTOCOLUMN_HISTORY_'
  '146 _REPEATCMD_HISTORY_'
  '152 _DOS_HISTORY_'
  '160 _FINDOPTIONS_HISTORY_'
  '161 _REPLACEOPTIONS_HISTORY_'
  '168 _FIND_HISTORY_'
  '169 _REPLACE_HISTORY_'
  '176 _FILLBLOCK_HISTORY_'
  '184 _HELP_SEARCH_HISTORY_'
end fixed_history_synonyms



// Global variables
string  cmd_fqn  [MAXSTRINGLEN] = ''
string  log_fqn  [MAXSTRINGLEN] = ''
integer ok                      = TRUE
integer started_from_main       = FALSE



/*
  search_path() works almost like SearchPath(), but if subdir is supplied:
  - Not searching the current directory.
  - Also searching TSE's program directory.
  For example, if subdir is supplied, and paths is "path1;path2", then these
  directories are searched in this order by search_path():
    path1
    path1\subdir
    path2
    path2\subdir
    LoadDir()
    LoadDir()\subdir
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)\subdir

  search_path_helper() should not be called directly.
  It is a helper proc specifically for use by search_path().
*/
string proc search_path_helper(string filename, string path , string subdir)
  string result [MAXSTRINGLEN] = ''

  if FileExists(path + filename)
    result = path + filename
  elseif subdir <> ''
  and    FileExists(path + subdir + SLASH + filename)
    result = path + subdir + SLASH + filename
  endif

  return(result)
end search_path_helper


string proc search_path(string filename, string paths, string subdir)
  integer i                          = 0
  string  path        [MAXSTRINGLEN] = ''
  string  program_dir [MAXSTRINGLEN] = ''
  string  result      [MAXSTRINGLEN] = ''

  for i = 1 to NumTokens(paths, ';')
    path   = GetToken(paths, ';', i)
    result = search_path_helper(filename, path + SLASH, subdir)
    if result <> ''
      break
    endif
  endfor

  if result == ''
    path   = LoadDir()
    result = search_path_helper(filename, path, subdir)
  endif

  if result == ''
    program_dir = SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    if program_dir <> path
      result = search_path_helper(filename, program_dir, subdir)
    endif
  endif

  return(result)
end search_path


string proc date_time_for_file_name(string yyyy_mm_dd, string hh_mm_ss)
  string uyyyymmdduhhmmss [16] = ''

  uyyyymmdduhhmmss = '_'             +
                     yyyy_mm_dd[1:4] +
                     yyyy_mm_dd[6:2] +
                     yyyy_mm_dd[9:2] +
                     '_'             +
                     hh_mm_ss  [1:2] +
                     hh_mm_ss  [4:2] +
                     hh_mm_ss  [7:2]

  // TSE 4.4 upwards supports TimeFormat 3, which already prefixes single-digit
  // hours with a zero instead of a space.
  #if EDITOR_VERSION < 4400h
    if uyyyymmdduhhmmss[11] == ' '
      uyyyymmdduhhmmss[11] = '0'
    endif
  #endif

  return(uyyyymmdduhhmmss)
end date_time_for_file_name


proc on_abandon_editor()
  integer confirmation = FALSE
  integer log_handle   = -1

  if MsgBox('Reminder',
         'Now that TSE closes, tsehist.dat will be repaired.',
         _YES_NO_) == 1
    confirmation = TRUE

    #ifdef LINUX
      // There is a bug in the Linux _RUN_DETACHED_ dos() call,
      // so we spawn a background process ourselves with "nohup ... &".
      Dos('nohup ' + cmd_fqn + ' > /dev/null 2>&1 &', DOS_SYNC_CALL_FLAGS)
    #else
      if isGUI()
        #if EDITOR_VERSION >= 4400h
          StartPgm(cmd_fqn, '', '', _START_HIDDEN_)
        #else
          Dos(cmd_fqn, DOS_ASYNC_CALL_FLAGS)
        #endif
      else
          Dos(cmd_fqn, DOS_ASYNC_CALL_FLAGS)
      endif
    #endif
  else
    EraseDiskFile(cmd_fqn)
    Warn(MACRO_NAME, ': tsehist repair aborted.')
  endif

  log_handle = fOpen(log_fqn, _OPEN_READWRITE_)
  if log_handle == -1
    Warn(MACRO_NAME, 'error: Could not open file'; QuotePath(log_fqn), ' (2).')
    ok = FALSE
  elseif fSeek(log_handle, 0, _SEEK_END_) == -1
    Warn(MACRO_NAME, 'error: Could not seek file'; QuotePath(log_fqn), ' (2).')
    ok = FALSE
  elseif fWrite(log_handle, Format('Session'; GetWinHandle(); 'at';
                                   GetDateStr(); GetTimeStr(), ':';
                                   'The previously accepted actions were';
                                   iif(confirmation, 'confirmed', 'aborted'),
                                   ' upon closing TSE.',
                                   NEWLINE, NEWLINE)) == -1
    Warn(MACRO_NAME, 'error: Could not write file';
         QuotePath(log_fqn), ' (2).')
    ok = FALSE
  elseif fClose(log_handle) == -1
    Warn(MACRO_NAME, 'error: Could not close file';
         QuotePath(log_fqn), ' (2).')
    ok = FALSE
  endif
end on_abandon_editor


proc prep_tsehist_update(integer hst_id,
                         string  tsehist_fqn,
                         integer num_history_names_new,
                         integer cmd_id)
  integer old_DateFormat                 = Set(DateFormat, 6)
  integer old_TimeFormat                 = Set(TimeFormat, NEW_TIMEFORMAT)
  string  tmp_dir         [MAXSTRINGLEN] = ''
  string  tsehist_bak_fqn [MAXSTRINGLEN] = ''
  string  tsehist_new_fqn [MAXSTRINGLEN] = ''

  if not FindThisFile(tsehist_fqn)
    Warn(MACRO_NAME; 'error: Could not access "', tsehist_fqn, '".')
    ok = FALSE
  endif

  if ok
    tsehist_bak_fqn = SplitPath(tsehist_fqn, _DRIVE_|_PATH_|_NAME_)       +
                      date_time_for_file_name( FFDateStr(),  FFTimeStr()) +
                      '_bak'                                              +
                      SplitPath(tsehist_fqn, _EXT_)
    tsehist_new_fqn = SplitPath(tsehist_fqn, _DRIVE_|_PATH_|_NAME_)       +
                      date_time_for_file_name(GetDateStr(), GetTimeStr()) +
                      '_new'                                              +
                      SplitPath(tsehist_fqn, _EXT_)
    if not FileExists(tsehist_bak_fqn)
      if not copy_file(tsehist_fqn, tsehist_bak_fqn)
        Warn(MACRO_NAME; 'error: Could not backup';
             QuotePath(tsehist_fqn); 'to'; QuotePath(tsehist_fqn), '.')
        ok = FALSE
      endif
    endif
  endif

  if ok
    GotoBufferId(hst_id)
    BegFile()
    InsertText(Chr(num_history_names_new), _OVERWRITE_)
    while lFind('^$', 'gx')
      KillLine()
    endwhile
    if not SaveAs(tsehist_new_fqn, _DONT_PROMPT_|_OVERWRITE_)
      Warn(MACRO_NAME; 'error: Could not save'; QuotePath(tsehist_new_fqn), '.')
      ok = FALSE
    endif
  endif

  if ok
    #ifdef LINUX
      tmp_dir = '/tmp'
    #else
      tmp_dir = RemoveTrailingSlash(GetEnvStr('TMP'))
      if tmp_dir == ''
        tmp_dir = RemoveTrailingSlash(GetEnvStr('TEMP'))
      endif
    #endif
    if not (FileExists(tmp_dir) & _DIRECTORY_)
      Warn(MACRO_NAME;
           'error: Cannot determine a temporary directory to use.')
      ok = FALSE
    endif
  endif

  if ok
    cmd_fqn = Format(tmp_dir, SLASH, 'Tse_', MACRO_NAME, '_', GetWinHandle(),
                     '.cmd')
    GotoBufferId(cmd_id)

    #ifdef LINUX
      AddLine(Format('#! /bin/bash'))
    #else
      AddLine(Format('echo on'))
    #endif

    // Pinging the local computer 2 times is a trick to wait about a second.
    // This delays the tsehist.dat update, making extra sure that TSE had time
    // to fully close and write its old tsehist.dat first.
    // Testing suggests that in practice the delay might not be necessary.
    AddLine(Format('ping -', PING_FREQUENCY_OPTION,' 2 127.0.0.1'))

    // In Windows I want to end with an option to avoid a "double quote at end
    // of line" exception in the Windows command interpreter.
    AddLine(Format(COPY_COMMAND; QuotePath(tsehist_new_fqn);
                   QuotePath(tsehist_fqn), COPY_OPTION))
    AddLine(Format(DELETE_COMMAND; QuotePath(cmd_fqn), DELETE_OPTION))
  endif

  if ok
    if SaveAs(cmd_fqn, _DONT_PROMPT_|_OVERWRITE_)
      #ifdef LINUX
        Dos('chmod ug+rwx ' + cmd_fqn, DOS_SYNC_CALL_FLAGS)
      #endif
    else
      Warn(MACRO_NAME; 'error: Cannot save'; QuotePath(cmd_fqn), '.')
      ok = FALSE
    endif
  endif

  if ok
    Hook(_ON_ABANDON_EDITOR_, on_abandon_editor)
    BegFile()
    UpdateDisplay()
    MsgBox('Prepared', 'After TSE closes, tsehist.dat will be repaired.')
  endif

  Set(DateFormat , old_DateFormat )
  Set(TimeFormat , old_TimeFormat )
end prep_tsehist_update


Keydef multiple_name_list_keys
  <Del> PushKey(<Enter>)
end multiple_name_list_keys


proc multiple_name_list_cleanup()
  Disable(multiple_name_list_keys)
end multiple_name_list_cleanup


proc multiple_name_list_startup()
  UnHook(multiple_name_list_startup)
  Enable(multiple_name_list_keys)
  Hook(_LIST_CLEANUP_, multiple_name_list_cleanup)
  ListFooter('<Del> Delete history list   <Enter> Select 1 name, delete other names')
end multiple_name_list_startup


string proc get_formatted_history_synonym(integer history_list_number,
                                          integer synonyms_id)
  string formatted_name [MAXSTRINGLEN] = ''
  PushLocation()
  GotoBufferId(synonyms_id)
  if not NumLines()
    PushBlock()
    InsertData(fixed_history_synonyms)
    PopBlock()
    BegFile()
    repeat
      InsertText(Chr(Val(GetText(1, 3))) + ' ', _INSERT_)
      BegLine()
    until not Down()
  endif
  if lFind('^' + Chr(history_list_number), 'gx')
    formatted_name = ' (' + GetToken(GetText(1, MAXSTRINGLEN), ' ', 3) + ')'
  else
    formatted_name = ' (unknown)'
  endif
  PopLocation()
  return(formatted_name)
end get_formatted_history_synonym


forward proc idle()


proc check_tsehist()
  integer actions_id                 = 0
  integer cmd_id                     = 0
  integer consent                    = FALSE
  integer deletions                  = 0
  integer double_name_intro_given    = FALSE
  integer empty_lines                = 0
  string  hist_name   [MAXSTRINGLEN] = ''
  string  hist_names  [MAXSTRINGLEN] = ''
  integer hist_num                   = 0
  integer hst_id                     = 0
  integer log_handle                 = -1
  integer names_id                   = 0
  integer num_history_names_new      = 0
  integer num_history_names_old      = 0
  integer num_hist_0_lines           = 0
  integer old_CenterFinds            = Set(CenterFinds, ON)
  string  search_options        [11] = ''
  integer selected_line              = 0
  integer synonyms_id                = 0
  string  tsehist_fqn [MAXSTRINGLEN] = ''
  integer values_id                  = 0

  UnHook(idle)

  if ok
    tsehist_fqn = search_path('tsehist.dat', '', '')
    if tsehist_fqn == ''
      Warn(MACRO_NAME; 'error: File "tsehist.dat" not found.')
      ok = FALSE
    endif
  endif

  if ok
    PushLocation()

    actions_id  = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':Actions',
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    cmd_id      = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':Cmd',
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    hst_id      = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':TseHist',
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    names_id    = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':ListNames',
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    synonyms_id = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':FixedHistorySynonyms',
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    values_id   = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + " - This history list's values",
                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)

    GotoBufferId(hst_id)

    if not LoadBuffer(TSEHIST_FQN, -2)
      Warn(MACRO_NAME; 'error: Could not load "', TSEHIST_FQN, '".')
      ok = FALSE
    endif
  endif

  if ok
    PushBlock()
    BegFile()
    num_history_names_old = CurrChar()
    num_history_names_new = num_history_names_old

    if num_history_names_old
      MarkLine(2, num_history_names_old + 1) // Mark the name lines.
    else
      UnMarkBlock()
    endif

    // Check error: Empty file lines. (Not same as empty history list lines.)
    search_options = 'gx'
    while lFind('^$', search_options)
      if isCursorInBlock()
        num_history_names_new = num_history_names_new - 1
      endif
      empty_lines    = empty_lines + 1
      search_options = 'x+'
    endwhile
    if empty_lines
      AddLine(Format('Delete'; empty_lines; 'empty lines.'), actions_id)
    endif

    // Check error: A "name" for a fixed history list.
    while lFind('^[\d128-\d255]', 'glx')
      num_history_names_new = num_history_names_new - 1
      AddLine(Format('Delete "name" for fixed history list'; CurrChar(),
                     get_formatted_history_synonym(CurrChar(), synonyms_id),
                     ': "', GetText(2, MAXSTRINGLEN), '".'),
              actions_id)
      KillToEol() // Later delete as empty line.
    endwhile

    // Check error: Free history list with number 0.
    BegFile()
    while lFind('^\d000', 'x+')
      if isCursorInBlock()
        num_history_names_new = num_history_names_new - 1
      endif
      KillToEol() // Later delete as empty line.
      num_hist_0_lines = num_hist_0_lines + 1
    endwhile
    if num_hist_0_lines
      AddLine(Format('Delete the'; num_hist_0_lines;
                     'lines of free history list 0.'),
              actions_id)
    endif

    // Check error: Free history lists with no values.
    if num_history_names_old
      for hist_num = 1 to 127
        GotoLine(num_history_names_old + 2) // Go to first value line.
        BegLine()
        if not lFind('^\d' + Format(hist_num:3:'0'), 'x')
          hist_names = ''
          while lFind('^\d' + Format(hist_num:3:'0'), 'glx')
            num_history_names_new = num_history_names_new - 1
            if hist_names <> ''
              hist_names = hist_names + ' and '
            endif
            hist_names = hist_names + '"' + GetText(2, MAXSTRINGLEN) + '"'
            KillToEol()
          endwhile
          if hist_names <> ''
            AddLine(Format('Delete empty free history list'; hist_num:3,
                           ' named'; hist_names, '.'),
                    actions_id)
          endif
        endif
      endfor
    endif

    // Check error: Free history lists with no name.
    // Note: This one is theoretical. I have not seen it occur.
    for hist_num = 1 to 127
      GotoLine(num_history_names_old + 2)   // Go to first value line.
      BegLine()
      if      lFind('^\d' + Format(hist_num:3:'0'), 'x')
      and not lFind('^\d' + Format(hist_num:3:'0'), 'glx')
        deletions = 0
        while lFind('^\d' + Format(hist_num:3:'0'), 'gx')
          deletions = deletions + 1
          KillToEol()
        endwhile
        AddLine(Format('Delete the'; deletions;
                       'values of undefined free history list';
                       hist_num, '.'),
                actions_id)
      endif
    endfor

    // Check error: History lists with multiple names.
    if num_history_names_old
      for hist_num = 1 to 127
        GotoBufferId(names_id)
        EmptyBuffer()
        GotoBufferId(hst_id)
        search_options = 'glx'
        while lFind(Format('^\d', hist_num:3:'0'), search_options)
          AddLine(Format(CurrLine():10, '   ', hist_num:3, '   "',
                         GetText(2, MAXSTRINGLEN), '"'),
                  names_id)
          search_options = 'lx+'
        endwhile
        GotoBufferId(names_id)
        if NumLines() >= 2
          if not double_name_intro_given
            MsgBox('Info',
                   Format(MACRO_NAME;
                          'found a free history list with more than one name.',
                          Chr(13), Chr(13),
                          'Next: Select one name for it, or delete the whole history list.'))
            double_name_intro_given = TRUE
          endif
          // Show the free the list's values in the background.
          GotoBufferId(values_id)
          EmptyBuffer()
          GotoBufferId(hst_id)
          GotoLine(num_history_names_old + 2)
          BegLine()
          search_options = 'x'
          while lFind(Format('^\d', hist_num:3:'0'), search_options)
            AddLine(GetText(2, MAXSTRINGLEN), values_id)
            search_options = 'x+'
          endwhile
          GotoBufferId(values_id)
          BegFile()
          UpdateDisplay()
          GotoBufferId(names_id)
          BegFile()
          Hook(_LIST_STARTUP_, multiple_name_list_startup)
          if List(Format('      Line  List   The multiple names for this list':-70), 70)
            if Query(LastKey) == <Del>
              GotoBufferId(hst_id)
              hist_names = ''
              while lFind(Format('^\d', hist_num:3:'0'), 'gx')
                if isCursorInBlock()
                  num_history_names_new = num_history_names_new - 1
                  if hist_names <> ''
                    hist_names = hist_names + ' and '
                  endif
                  hist_names = hist_names + '"' + GetText(2, MAXSTRINGLEN) + '"'
                endif
                KillToEol()
              endwhile
              if hist_names == ''
                hist_names = '.'
              else
                hist_names = ' named ' + hist_names + '.'
              endif
              AddLine(Format('Delete free history list';
                             hist_num:3, hist_names),
                      actions_id)
            else
              selected_line = Val(GetText(1, 10))
              GotoBufferId(hst_id)
              GotoLine(selected_line)
              hist_name      = GetText(2, MAXSTRINGLEN) // xx
              hist_names     = ''
              search_options = 'glx'
              while lFind(Format('^\d', hist_num:3:'0'), search_options)
                if CurrLine() <> selected_line
                  num_history_names_new = num_history_names_new - 1
                  if hist_names <> ''
                    hist_names = hist_names + ' and '
                  endif
                  hist_names = hist_names + '"' + GetText(2, MAXSTRINGLEN) + '"'
                  KillToEol()
                endif
                search_options = 'lx+'
              endwhile
              AddLine(Format('For free history list'; hist_num:3; ' "',
                             hist_name, '"'; 'delete alternate name ',
                             hist_names, '.'),
                      actions_id)
            endif
          endif
        endif
        GotoBufferId(hst_id)
      endfor
    endif

    PopBlock()

    GotoBufferId(actions_id)
    if NumLines()
      BegFile()
      UpdateDisplay()
      // We are popping up the list during editing, so first empty the keyboard
      // buffer to stop and ignore a user('s) typing.
      while KeyPressed()
      or    KeyPressed()  // Bouble KeyPressed() for bug in old TSE versions.
        Alarm()
        GetKey()
        Delay(9)
      endwhile
      // The list might be so large as to be confusing.
      // So first give a simple intro.
      if  MsgBox('Info',
                 Format(MACRO_NAME;
                        'found errors in TSE history lists.',
                        Chr(13), Chr(13),
                        'Press <Enter> to list and repair them, or <Escape> to quit.'),
                 _OK_)
            == 1
      and List('Press <Enter> to do all these history repair actions, or <Escape> to quit',
               Max(73, LongestLineInBuffer()))
        consent = TRUE
        prep_tsehist_update(hst_id, tsehist_fqn, num_history_names_new, cmd_id)
      else
        MsgBox('Info',
               Format('You escaped!', Chr(13), Chr(13),
                      'Erroneous history lists will NOT be repaired.'))

      endif

      GotoBufferId(actions_id)

      log_fqn = SplitPath(tsehist_fqn, _DRIVE_|_PATH_|_NAME_) + '_' +
                MACRO_NAME + '.log'
      if FileExists(log_fqn)
        log_handle = fOpen(log_fqn, _OPEN_READWRITE_)
      else
        log_handle = fCreate(log_fqn)
      endif
      if log_handle == -1
        Warn(MACRO_NAME, 'error: Could not open file'; QuotePath(log_fqn), ' (1).')
        ok = FALSE
      elseif fSeek(log_handle, 0, _SEEK_END_) == -1
        Warn(MACRO_NAME, 'error: Could not seek file'; QuotePath(log_fqn), ' (1).')
        ok = FALSE
      elseif fWrite(log_handle,
                    Format('Session'; GetWinHandle();
                           'at'; GetDateStr(); GetTimeStr(), ':';
                           'Proposed actions:',
                           NEWLINE, NEWLINE)) == -1
        Warn(MACRO_NAME, 'error: Could not write file';
             QuotePath(log_fqn), ' (1).')
        ok = FALSE
      elseif not fWriteFile(log_handle, _DONT_PROMPT_|_APPEND_)
        Warn(MACRO_NAME, 'error: Could not append to file'; QuotePath(log_fqn), ' (1).')
        ok = FALSE
      elseif fWrite(log_handle,
                    Format(NEWLINE,
                           'Session'; GetWinHandle();
                           'at'; GetDateStr(); GetTimeStr(), ':';
                           'The proposed actions were';
                           iif(consent, 'accepted', 'rejected'), '.',
                           NEWLINE, NEWLINE)) == -1
        Warn(MACRO_NAME, 'error: Could not write file';
             QuotePath(log_fqn), ' (1).')
        ok = FALSE
      elseif fClose(log_handle) == -1
        Warn(MACRO_NAME, 'error: Could not close file';
             QuotePath(log_fqn), ' (1).')
        ok = FALSE
      endif
    else
      if started_from_main
        Warn(MACRO_NAME, ': Nothing to do.')
      endif

      PurgeMacro(MACRO_NAME)
    endif
  endif

  PopLocation()
  AbandonFile(actions_id)
  AbandonFile(cmd_id)
  AbandonFile(hst_id)
  AbandonFile(names_id)
  AbandonFile(synonyms_id)
  AbandonFile(values_id)
  Set(CenterFinds, old_CenterFinds)
  UpdateDisplay()

  if not ok
    PurgeMacro(MACRO_NAME)
  endif
end check_tsehist


proc idle()
  check_tsehist()
end idle


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)

  Hook(_IDLE_, idle)
end WhenLoaded


proc Main()
  started_from_main = TRUE
  check_tsehist()
end Main

